# Licensed under the MIT License - https://opensource.org/licenses/MIT

from sklearn import neighbors, tree, svm
from sklearn.linear_model import SGDClassifier, LogisticRegression
from sklearn.utils import shuffle
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.naive_bayes import GaussianNB
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.neural_network import MLPClassifier

import math
import numpy as np
import random
import logging
import numbers


logger = logging.getLogger('pycobra.classifiercobra')


class ClassifierCobra(BaseEstimator):
    """
    Classification algorithm as introduced by
    Mojirsheibani [1999] Combining Classifiers via Discretization,
    Journal of the American Statistical Association.

    Parameters
    ----------
    random_state: integer or a numpy.random.RandomState object.
        Set the state of the random number generator to pass on to shuffle and loading machines, to ensure
        reproducibility of your experiments, for example.

    Attributes
    ----------
    machines: A dictionary which maps machine names to the machine objects.
            The machine object must have a predict method for it to be used during aggregation.
    machine_predictions: A dictionary which maps machine name to it's predictions over X_l
            This value is used to determine which points from y_l are used to aggregate.

    """
    def __init__(self, random_state=None, machine_list='basic'):
        self.random_state = random_state
        self.machine_list = machine_list

    def fit(self, X, y, default=True, X_k=None, X_l=None, y_k=None, y_l=None):
        """
        Parameters
        ----------
        X: array-like, [n_samples, n_features]
            Training data which will be used to create ClassifierCobra.

        y: array-like [n_samples]
            Training labels for classification.

        default: bool, optional
            If set as true then sets up COBRA with default machines and splitting.

        X_k : shape = [n_samples, n_features]
            Training data which is used to train the machines loaded into COBRA.

        y_k : array-like, shape = [n_samples]
            Target values used to train the machines loaded into COBRA.

        X_l : shape = [n_samples, n_features]
            Training data which is used during the aggregation of COBRA.

        y_l : array-like, shape = [n_samples]
            Target values which are actually used in the aggregation of COBRA.
        """

        X, y = check_X_y(X, y)
        self.X_ = X
        self.y_ = y
        self.X_k_ = X_k
        self.X_l_ = X_l
        self.y_k_ = y_k
        self.y_l_ = y_l
        self.estimators_ = {}

        # try block to pass scikit-learn estimator check.
        try:
            # set-up COBRA with default machines
            if default:
                self.split_data()
                self.load_default(machine_list=self.machine_list)
                self.load_machine_predictions()
        except ValueError:
            return self

        return self


    def pred(self, X, M, info=False):
        """
        Performs the CLassififerCobra aggregation scheme, used in predict method.

        Parameters
        ----------
        X: array-like, [n_features]

        M: int, optional
            M refers to the number of machines the prediction must be close to to be considered during aggregation.

        info: boolean, optional
            If info is true the list of points selected in the aggregation is returned.

        Returns
        -------
        result: prediction

        """

        # dictionary mapping machine to points selected
        select = {}
        for machine in self.estimators_:
            # machine prediction
            label = self.estimators_[machine].predict(X)
            select[machine] = set()
            # iterating from l to n
            # replace with numpy iteration
            for count in range(0, len(self.X_l_)):
                if self.machine_predictions_[machine][count] == label:
                    select[machine].add(count)

        points = []
        # count is the indice number.
        for count in range(0, len(self.X_l_)):
            # row check is number of machines which picked up a particular point
            row_check = 0
            for machine in select:
                if count in select[machine]:
                    row_check += 1
            if row_check == M:
                points.append(count)

        # if no points are selected, return 0
        if len(points) == 0:
            if info:
                logger.info("No points were selected, prediction is 0")
                return (0, 0)
            logger.info("No points were selected, prediction is 0")
            return 0

        # aggregate
        classes = {}
        for label in np.unique(self.y_l_):
            classes[label] = 0

        for point in points:
            classes[self.y_l_[point]] += 1

        result = int(max(classes, key=classes.get))
        if info:
            return result, points
        return result


    def predict(self, X, M=None, info=False):
        """
        Performs the ClassifierCobra aggregation scheme, calls pred.
        ClassifierCobra performs a majority vote among all points which are retained by the COBRA procedure.

        Parameters
        ----------
        X: array-like, [n_features]

        M: int, optional
            M refers to the number of machines the prediction must be close to to be considered during aggregation.

        info: boolean, optional
            If info is true the list of points selected in the aggregation is returned.

        Returns
        -------
        result: prediction

        """
        X = check_array(X)

        if M is None:
            M = len(self.estimators_)
        if X.ndim == 1:
            return self.pred(X.reshape(1, -1), M=M)

        result = np.zeros(len(X))
        avg_points = 0
        index = 0
        for vector in X:
            if info:
                result[index], points = self.pred(vector.reshape(1, -1), M=M, info=info)
                avg_points += len(points)
            else:
                result[index] = self.pred(vector.reshape(1, -1), M=M)
            index += 1
        
        if info:
            avg_points = avg_points / len(X_array)
            return result, avg_points
        
        return result


    def predict_proba(self, X, kernel=None, metric=None, bandwidth=1, **kwargs): 
        """
        Performs the ClassifierCobra aggregation scheme and calculates probability of a point being in a particular class.
        ClassifierCobra performs a majority vote among all points which are retained by the COBRA procedure.
        
        NOTE: this method is to visualise boundaries.
        The current method is just the mean of the consituent machines, as the concept of that kind of predicted probability
        doesn't exist (yet) for classifier cobra.

        Parameters
        ----------
        X: array-like, [n_features]
        """

        probs = []
        for machine in self.estimators_:
            try:
                probs.append(self.estimators_[machine].predict_proba(X))
            except AttributeError:
                continue
        prob = np.mean(probs, axis=0)
        return prob


    def split_data(self, k=None, l=None, shuffle_data=True):
        """
        Split the data into different parts for training machines and for aggregation.

        Parameters
        ----------
        k : int, optional
            k is the number of points used to train the machines.
            Those are the first k points of the data provided.

        l: int, optional
            l is the number of points used to form the ClassifierCobra aggregate.

        shuffle: bool, optional
            Boolean value to decide to shuffle the data before splitting.

        Returns
        -------
        self : returns an instance of self.
        """

        if shuffle_data:
            self.X_, self.y_ = shuffle(self.X_, self.y_, random_state=self.random_state)

        if k is None and l is None:
            k = int(len(self.X_) / 2)
            l = int(len(self.X_))

        if k is not None and l is None:
            l = len(self.X_) - k

        if l is not None and k is None:
            k = len(self.X_) - l

        self.X_k_ = self.X_[:k]
        self.X_l_ = self.X_[k:l]
        self.y_k_ = self.y_[:k]
        self.y_l_ = self.y_[k:l]

        return self


    def load_default(self, machine_list='basic'):
        """
        Loads 4 different scikit-learn regressors by default. The advanced list adds more machines. 
        As of current release SGD algorithm is not included in the advanced list.

        Parameters
        ----------
        machine_list: optional, list of strings
            List of default machine names to be loaded.
        Returns
        -------
        self : returns an instance of self.
        """
        if machine_list == 'basic':
            machine_list = ['sgd', 'tree', 'knn', 'svm']
        if machine_list == 'advanced':
            machine_list = ['tree', 'knn', 'svm', 'logreg', 'naive_bayes', 'lda', 'neural_network']

        for machine in machine_list:
            try:
                if machine == 'svm':
                    self.estimators_['svm'] = svm.SVC().fit(self.X_k_, self.y_k_)
                if machine == 'knn':
                    self.estimators_['knn'] = neighbors.KNeighborsClassifier().fit(self.X_k_, self.y_k_)
                if machine == 'tree':
                    self.estimators_['tree'] = tree.DecisionTreeClassifier().fit(self.X_k_, self.y_k_)
                if machine == 'logreg':
                    self.estimators_['logreg'] = LogisticRegression(random_state=self.random_state).fit(self.X_k_, self.y_k_)
                if machine == 'naive_bayes':
                    self.estimators_['naive_bayes'] = GaussianNB().fit(self.X_k_, self.y_k_)
                if machine == 'lda':
                    self.estimators_['lda'] = LinearDiscriminantAnalysis().fit(self.X_k_, self.y_k_)
                if machine == 'neural_network':
                    self.estimators_['neural_network'] = MLPClassifier(random_state=self.random_state).fit(self.X_k_, self.y_k_)
            except ValueError:
                continue

        return self


    def load_machine(self, machine_name, machine):
        """
        Adds a machine to be used during the aggregation strategy.
        The machine object must have been trained using X_k and y_k, and must have a 'predict()' method.
        After the machine is loaded, for it to be used during aggregation, load_machine_predictions must be run.

        Parameters
        ----------
        machine_name : string
            Name of the machine you are loading

        machine: machine/regressor object
            The regressor machine object which is mapped to the machine_name

        Returns
        -------
        self : returns an instance of self.
        """

        self.estimators_[machine_name] = machine

        return self


    def load_machine_predictions(self, predictions=None):
        """
        Stores the trained machines' predicitons on D_l in a dictionary, to be used for predictions.
        Should be run after all the machines to be used for aggregation is loaded.

        Parameters
        ----------
        predictions: dictionary, optional
            A pre-existing machine:predictions dictionary can also be loaded.

        Returns
        -------
        self : returns an instance of self.
        """
        self.machine_predictions_ = {}
        if predictions is None:
            for machine in self.estimators_:
                self.machine_predictions_[machine] = self.estimators_[machine].predict(self.X_l_)

        return self


    def load_machine_proba_predictions(self, predictions=None):
        """
        Stores the trained machines' predicitons on D_l in a dictionary, to be used for predictions.
        Should be run after all the machines to be used for aggregation is loaded.

        Parameters
        ----------
        predictions: dictionary, optional
            A pre-existing machine:predictions dictionary can also be loaded.

        Returns
        -------
        self : returns an instance of self.
        """
        self.machine_proba_predictions_ = {}
        if predictions is None:
            for machine in self.estimators_:
                try:
                    self.machine_proba_predictions_[machine] = self.estimators_[machine].predict_proba(self.X_l_)
                except AttributeError:
                    self.machine_proba_predictions_[machine] = self.estimators_[machine].decision_function(self.X_l_)
        return self




# Licensed under the MIT License - https://opensource.org/licenses/MIT
import math
import itertools

import numpy as np
import pandas as pd
import seaborn as sns

import matplotlib.pyplot as plt
import matplotlib.cm as cm
import random
from scipy.spatial import Voronoi, voronoi_plot_2d
from sklearn.metrics import mean_squared_error, accuracy_score
from sklearn.utils import shuffle

from pycobra.diagnostics import Diagnostics
from pycobra.cobra import Cobra
from pycobra.kernelcobra import KernelCobra

from pycobra.ewa import Ewa
from pycobra.classifiercobra import ClassifierCobra

from collections import OrderedDict

import logging

logger = logging.getLogger('pycobra.visualisation')


def create_labels(indice_info):
    """
    Helper method to create labels for plotting.

    Parameters
    ----------

    indice_info: list of strings
        List of machine names

    Return
    ------

    label: string
        Serves as a label during plotting.
    """

    label = ""
    for machine in indice_info:
        label = machine + " + " + label
    return label[:-3]


def gen_machine_colors(only_colors=False, num_colors=None, indice_info=None, rgb=False, plot_machines=None, colors=None):
    """
    Helper method to create a machine combinations to color dictionary, or a list of colors.

    Parameters
    ----------

    indice_info: dictionary, optional
        Dictionary which is a result of running pycobra.visualisation.indice_info. Maps indices to combinations of machines.

    only_colors: bool, optional
        Option to return only a list of colors

    num_colors: int, optional
        Number of colors to be returned if using only_colors

    rgb : bool, optional
        Creates dictionary based on machine used and r, g, b, a scheme.

    plot_machines: list of strings, optional
        List of machines to use in rgb coloring.

    colors: list of strings, optional
        List of colors to be used for pairing with machine_combinations

    Return
    ------

    machine_colors: dictionary
        Dictionary mapping machine combinations and color.
    """

    # note: for default colors to be assigned, the latest version of matplotlib is needed.
    # the code below is taken from the colors example:
    from matplotlib import colors as mcolors
    colors = dict(mcolors.BASE_COLORS, **mcolors.CSS4_COLORS)
    # Sort colors by hue, saturation, value and name.
    by_hsv = sorted((tuple(mcolors.rgb_to_hsv(mcolors.to_rgba(color)[:3])), name)
                    for name, color in colors.items())
    sorted_names = [name for hsv, name in by_hsv]

    # if we need only as many colors as individual machine
    if only_colors:
        # make the sorted names list equal to in size as the machine combinations
        # we do this by first picking up equally spaced out colors, then randonly picking out the right amount
        sorted_names = sorted_names[0::int(len(sorted_names) / num_colors)]
        return random.sample(sorted_names, num_colors)

    machine_combinations = list(set(indice_info.values()))
    machine_colors = {}

    if rgb:
        for indice in indice_info:
            r, g, b, a = 0, 0, 0, 0.4
            if plot_machines[0] in indice_info[indice]:
                r = 1
            if plot_machines[1] in indice_info[indice]:
                g = 1
            if plot_machines[2] in indice_info[indice]:
                b = 1
            if plot_machines[3] in indice_info[indice]:
                a = 1
            if (r, g, b) == (1, 1, 1):
                r, g, b = 0, 0, 0
            machine_colors[indice_info[indice]] = (r, g, b, a)
        return machine_colors

    # if it isn't rgb, let's pair each unique machine with a color provided
    if colors is not None and len(machine_combinations) == len(colors):
        for machine, color in zip(machine_combinations, colors):
            machine_colors[machine] = color
        return machine_colors

    # if it's none of the above options, we create colors similar to the only colors option.
    sorted_names = sorted_names[0::int(len(sorted_names) / len(machine_combinations))]
    colors = random.sample(sorted_names, len(machine_combinations))

    for machine, color in zip(machine_combinations, colors):
        machine_colors[machine] = color
    return machine_colors


def voronoi_finite_polygons_2d(vor, radius=None):
    """
    Code originally written by pv: https://gist.github.com/pv/8036995.
    Helper method for voronoi.
    Reconstruct infinite voronoi regions in a 2D diagram to finite
    regions.

    Parameters
    ----------
    vor : Voronoi
        Input diagram
    radius : float, optional
        Distance to 'points at infinity'.

    Returns
    -------
    regions : list of tuples
        Indices of vertices in each revised Voronoi regions.
    vertices : list of tuples
        Coordinates for revised Voronoi vertices. Same as coordinates
        of input vertices, with 'points at infinity' appended to the
        end.

    """

    if vor.points.shape[1] != 2:
        raise ValueError("Requires 2D input")

    new_regions = []
    new_vertices = vor.vertices.tolist()

    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max()*2

    # Construct a map containing all ridges for a given point
    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))

    # Reconstruct infinite regions
    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]

        if all([v >= 0 for v in vertices]):
            # finite region
            new_regions.append(vertices)
            continue

        # reconstruct a non-finite region
        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]

        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                # finite ridge: already in the region
                continue

            # Compute the missing endpoint of an infinite ridge
            t = vor.points[p2] - vor.points[p1]  # tangent
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])  # normal

            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius

            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())

        # sort region counterclockwise
        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:, 1] - c[1], vs[:, 0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]

        # finish
        new_regions.append(new_region.tolist())

    return new_regions, np.asarray(new_vertices)


class Visualisation():
    """
    Plots and visualisations of COBRA aggregates.
    If X_test and y_test is loaded, you can run the plotting functions with no parameters.
    """
    def __init__(self, aggregate, X_test, y_test, plot_size=8, estimators={}, random_state=None, **kwargs):
        """
        Parameters
        ----------
        aggregate: pycobra.cobra.Cobra or pycobra.cobra.Ewa object
            aggregate on which we want to run our analysis on.

        X_test : array-like, shape = [n_samples, n_features].
            Testing data.

        y_test : array-like, shape = [n_samples].
            Test data target values.

        plot_size: int, optional
            Size of matplotlib plots.

        estimators: list, optional
            List of machine objects to visualise. Default is machines used in aggregate.

        """
        self.aggregate = aggregate
        self.X_test = X_test
        self.y_test = y_test
        self.plot_size = plot_size
        self.estimators = estimators
        # load results so plotting doesn't need parameters
        self.kwargs = kwargs
        self.machine_test_results = {}
        self.machine_MSE = {}
        
        if len(self.estimators) == 0:
            self.estimators = self.aggregate.estimators_

        # if we are visualising ClassifierCobra then we must use accuracy score instead of MSE
        if type(aggregate) is ClassifierCobra:
            self.machine_test_results["ClassifierCobra"] = self.aggregate.predict(self.X_test)
            self.machine_error["ClassifierCobra"] = 1 - accuracy_score(self.y_test, self.machine_test_results["ClassifierCobra"])
            for machine in self.estimators_:
                self.machine_test_results[machine] = self.estimators_[machine].predict(self.X_test)
                # add MSE
                self.machine_error[machine] = 1 - accuracy_score(self.y_test, self.machine_test_results[machine])


        names_dict = {Cobra: "Cobra", Ewa: "EWA", KernelCobra: "KernelCobra"}
        for name in names_dict:
            if type(aggregate) == name:
                if type(aggregate) == KernelCobra:
                    self.machine_test_results[names_dict[name]] = self.aggregate.predict(self.X_test, bandwidth=kwargs["bandwidth_kernel"])
                else:
                    self.machine_test_results[names_dict[name]] = self.aggregate.predict(self.X_test)
                self.machine_MSE[names_dict[name]] = mean_squared_error(self.y_test, self.machine_test_results[names_dict[name]])            

        for machine in self.estimators:
            if type(self.estimators[machine]) == KernelCobra:
                self.machine_test_results[machine] = self.estimators[machine].predict(self.X_test, bandwidth=kwargs["bandwidth_kernel"])
            else:
                self.machine_test_results[machine] = self.estimators[machine].predict(self.X_test)
            self.machine_MSE[machine] = mean_squared_error(self.y_test, self.machine_test_results[machine])


        self.random_state = random_state
        if self.random_state is None:
            self.random_state = self.aggregate.random_state


    def plot_machines(self, machines=None, colors=None, plot_indices=False):
        """
        Plot the results of the machines versus the actual answers (testing space).

        Parameters
        ----------
        machines: list, optional
            List of machines to plot.
        colors: list, optional
            Colors of machines.
        plot_indices: boolean, optional.
            Plots truth values against indices.

        """

        if machines is None:
            machines = self.estimators

        plt.figure(figsize=(self.plot_size, self.plot_size))

        if plot_indices or self.X_test.size != self.y_test.size:
            linspace = np.linspace(0, len(self.y_test), len(self.y_test))

        if colors is None:
            colors = gen_machine_colors(only_colors=True, num_colors=len(machines) + 1)

        if plot_indices or self.X_test.size != self.y_test.size:
            plt.scatter(linspace, self.y_test, color=colors[0], label="Truth")
        else:
            plt.scatter(self.X_test, self.y_test, color=colors[0], label="Truth")

        for machine, color in zip(machines, colors[1:]):
            if plot_indices or self.X_test.size != self.y_test.size:
                plt.scatter(linspace, self.machine_test_results[machine], color=color, label=machine)
            else:
                plt.scatter(self.X_test, self.machine_test_results[machine], color=color, label=machine)

        if plot_indices:
            plt.xlabel("Point Indice")

        plt.legend()
        plt.show()
        return plt


    def QQ(self, machine="Cobra"):
        """
        Plots the machine results vs the actual results in the form of a QQ-plot.

        Parameters
        ----------
        machine: string, optional
            Name of machine to perform QQ-plot.
        """
        plt.figure(figsize=(self.plot_size, self.plot_size))
        axes = plt.gca()
        pred = self.machine_test_results[machine]

        # this is to make the plot look neater
        min_limits = math.fabs(min(min(pred), min(self.y_test)))
        max_limits = max(max(pred), max(self.y_test))
        axes.set_xlim([min(min(pred), min(self.y_test)) - min_limits, max(max(pred), max(self.y_test)) + max_limits])
        axes.set_ylim([min(min(pred), min(self.y_test)) - min_limits, max(max(pred), max(self.y_test)) + max_limits])

        # scatter the machine responses versus the actual y_test
        plt.scatter(self.y_test, pred, label=machine)
        axes.plot(axes.get_xlim(), axes.get_ylim(), ls="--", c=".3")

        # labels
        plt.xlabel('RESPONSES')
        plt.ylabel('PREDICTED')

        plt.legend()
        plt.show()
        return plt


    def boxplot(self, reps=100, info=False, dataframe=None, kind="normal"):
        """
        Plots boxplots of machines.

        Parameters
        ----------
        reps: int, optional
            Number of times to repeat experiments for boxplot.

        info: boolean, optional
            Returns data 

        """

        kwargs = self.kwargs
        if dataframe is None:
            if type(self.aggregate) is Cobra:

                MSE = {k: [] for k, v in self.estimators.items()}
                MSE["Cobra"] = []
                for i in range(0, reps):
                    cobra = Cobra(epsilon=self.aggregate.epsilon)
                    X, y = shuffle(self.aggregate.X_, self.aggregate.y_)
                    cobra.fit(X, y, default=False)
                    cobra.split_data(shuffle_data=True)

                    for machine in self.aggregate.estimators_:
                        self.aggregate.estimators_[machine].fit(cobra.X_k_, cobra.y_k_)
                        cobra.load_machine(machine, self.aggregate.estimators_[machine])

                    cobra.load_machine_predictions()

                    for machine in self.estimators:
                        if "Cobra" in machine:
                            self.estimators[machine].fit(X, y)
                        else:
                            self.estimators[machine].fit(cobra.X_k_, cobra.y_k_)
                        try:
                            if type(self.estimators[machine]) == KernelCobra:
                                preds = self.estimators[machine].predict(self.X_test, bandwidth=kwargs["bandwidth_kernel"])
                            else:
                                preds = self.estimators[machine].predict(self.X_test)
                        except KeyError:
                            preds = self.estimators[machine].predict(self.X_test)                      
                        
                        MSE[machine].append(mean_squared_error(self.y_test, preds))

                    MSE["Cobra"].append(mean_squared_error(self.y_test, cobra.predict(self.X_test)))

                try:
                    dataframe = pd.DataFrame(data=MSE)
                except ValueError:
                    return MSE

            if type(self.aggregate) is KernelCobra:

                MSE = {k: [] for k, v in self.estimators.items()}
                MSE["KernalCobra"] = []
                for i in range(0, reps):
                    kernel = KernelCobra()
                    X, y = shuffle(self.aggregate.X_, self.aggregate.y_)
                    kernel.fit(X, y, default=False)
                    kernel.split_data(shuffle_data=True)

                    for machine in self.aggregate.estimators_:
                        self.aggregate.estimators_[machine].fit(kernel.X_k_, kernel.y_k_)
                        kernel.load_machine(machine, self.aggregate.estimators_[machine])

                    kernel.load_machine_predictions()

                    for machine in self.estimators:
                        if "Cobra" in machine:
                            self.estimators[machine].fit(X, y)
                        else:
                            self.estimators[machine].fit(cobra.X_k_, cobra.y_k_)
                        
                        try:
                            if type(self.estimators[machine]) == KernelCobra:
                                preds = self.estimators[machine].predict(self.X_test, bandwidth=kwargs["bandwidth_kernel"])
                            else:
                                preds = self.estimators[machine].predict(self.X_test)
                        except KeyError:
                            preds = self.estimators[machine].predict(self.X_test)

                        MSE[machine].append(mean_squared_error(self.y_test, preds))

                    MSE["KernelCobra"].append(mean_squared_error(self.y_test, kernel.predict(self.X_test, bandwidth=kwargs[bandwidth_kernel])))

                try:
                    dataframe = pd.DataFrame(data=MSE)
                except ValueError:
                    return MSE


            if type(self.aggregate) is Ewa:

                MSE = {k: [] for k, v in self.aggregate.estimators_.items()}
                MSE["EWA"] = []
                for i in range(0, reps):
                    ewa = Ewa(random_state=self.random_state, beta=self.aggregate.beta)
                    X, y = shuffle(self.aggregate.X_, self.aggregate.y_, random_state=self.aggregate.random_state)
                    ewa.fit(X, y, default=False)
                    ewa.split_data(shuffle_data=True)

                    for machine in self.estimators:
                        self.aggregate.estimators_[machine].fit(ewa.X_k_, ewa.y_k_)
                        ewa.load_machine(machine, self.aggregate.estimators_[machine])

                    ewa.load_machine_weights(self.aggregate.beta)
                    X_test, y_test = shuffle(self.X_test, self.y_test, random_state=self.aggregate.random_state)
                    for machine in self.estimators:
                        if "EWA" in machine:
                            self.estimators[machine].fit(X, y)
                        else:
                            self.estimators[machine].fit(ewa.X_k_, ewa.y_k_)
                        try:
                            if type(self.estimators[machine]) == KernelCobra:
                                preds = self.estimators[machine].predict(self.X_test, bandwidth=kwargs["bandwidth_kernel"])
                            else:
                                preds = self.estimators[machine].predict(self.X_test)
                        except KeyError:
                            preds = self.estimators[machine].predict(self.X_test)                      
                        MSE[machine].append(mean_squared_error(y_test, preds))
                    
                    MSE["EWA"].append(mean_squared_error(y_test, ewa.predict(X_test)))

                try:
                    dataframe = pd.DataFrame(data=MSE)
                except ValueError:
                    return MSE

            if type(self.aggregate) is ClassifierCobra:

                errors = {k: [] for k, v in self.aggregate.estimators_.items()}
                errors["ClassifierCobra"] = []
                for i in range(0, reps):
                    cc = ClassifierCobra(random_state=self.random_state)
                    X, y = shuffle(self.aggregate.X_, self.aggregate.y_, random_state=self.aggregate.random_state)
                    cc.fit(X, y, default=False)
                    cc.split_data(shuffle_data=True)

                    for machine in self.aggregate.estimators_:
                        self.aggregate.estimators_[machine].fit(cc.X_k_, cc.y_k_)
                        cc.load_machine(machine, self.aggregate.estimators_[machine])

                    cc.load_machine_predictions()
                    X_test, y_test = shuffle(self.X_test, self.y_test, random_state=self.aggregate.random_state)
                    for machine in self.estimators: 
                        errors[machine].append(1 - accuracy_score(y_test, self.estimators[machine].predict(X_test)))
                    errors["ClassifierCobra"].append(1 - accuracy_score(y_test, cc.predict(X_test)))

                try:
                    dataframe = pd.DataFrame(data=errors)
                except ValueError:
                    return errors
        


        # code for different boxplot styles using the python graph gallery tutorial:
        # https://python-graph-gallery.com/39-hidden-data-under-boxplot/

        sns.set(style="whitegrid")

        if kind == "normal":
            sns.boxplot(data=dataframe)
            plt.title("Boxplot")

        if kind == "violin":
            sns.violinplot(data=dataframe)
            plt.title("Violin Plot")

        if kind == "jitterplot":
            ax = sns.boxplot(data=dataframe)
            ax = sns.stripplot(data=dataframe, color="orange", jitter=0.2, size=2.5)
            plt.title("Boxplot with jitter", loc="left")

        plt.ylabel("Mean Squared Errors")
        plt.xlabel("Estimators")
        plt.figure(figsize=(self.plot_size, self.plot_size))
        plt.show()

        
        if info:
            return dataframe

    def indice_info(self, X_test=None, y_test=None, epsilon=None, line_points=200):
        """
        Method to return information about each indices (query) optimal machines for testing data.

        Parameters
        ----------
        epsilon: float, optional
            Epsilon value to use for diagnostics

        line_points: int, optional
            if epsilon is not passed, optimal epsilon is found per point.

        Returns
        -------

        indice_info: dicitonary mapping indice to optimal machines.

        MSE: dictionary mapping indice to mean squared error for optimal machines for that point.

        """

        if X_test is None:
            X_test = self.X_test
        if y_test is None:
            y_test = self.y_test

        indice = 0
        indice_info = {}
        MSE = {}
        cobra_diagnostics = Diagnostics(aggregate=self.aggregate, random_state=self.random_state)
        if epsilon is None:
            for data_point, response in zip(X_test, y_test):
                info = cobra_diagnostics.optimal_machines_grid(data_point, response, line_points=line_points)
                indice_info[indice], MSE[indice] = info[0][0], info[1]
                indice += 1
        else:
            for data_point, response in zip(X_test, y_test):
                info = cobra_diagnostics.optimal_machines(data_point, response, single=True, epsilon=epsilon)
                indice_info[indice], MSE[indice] = info[0], info[1]
                indice += 1

        return indice_info, MSE


    def color_cobra(self, X_test=None, y_test=None, line_points=200, epsilon=None, indice_info=None, plot_machines=["ridge", "lasso", "random_forest", "tree"], single=False, machine_colors=None):
        """
        Plot the input space and color query points based on the optimal machine used for that point.

        Parameters
        ----------
        epsilon: float, optional
            Epsilon value to use for diagnostics. Used to find indice_info if it isn't passed.

        line_points: int, optional
            if epsilon is not passed, optimal epsilon is found per point. Used to find indice_info if it isn't passed.

        indice_info: dicitonary, optional
            dictionary mapping indice to optimal machines.

        plot_machines: list, optional
            list of machines to be plotted.

        single: bool, optional
            plots a single plot with each machine combination.

        machine_colors: dictionary, optional
            Depending on the kind of coloring, a dictionary mapping machines to colors.

        """

        if indice_info is None:
            indice_info = self.indice_info(line_points, epsilon)

        if X_test is None:
            X_test = self.X_test
        if y_test is None:
            y_test = self.y_test
        # we want to plot only two columns
        data_1 = X_test[:, 0]
        data_2 = X_test[:, 1]

        if single:
            if machine_colors is None:
                machine_colors = gen_machine_colors(indice_info=indice_info)
            plt.ion()
            fig, ax = plt.subplots()
            plot = ax.scatter([], [])
            for indice in indice_info:
       
                ax.set_title("All Machines")
                ax.scatter(data_1[indice], data_2[indice], color=machine_colors[indice_info[indice]], label=create_labels(indice_info[indice]))

            try:
                plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
            except ValueError:
                return ax

        if not single:
            if machine_colors is None:
                machine_colors = {}
                colors = gen_machine_colors(only_colors=True, num_colors=len(plot_machines))
                for machine, color in zip(plot_machines, colors):
                    machine_colors[machine] = color

            for machine in plot_machines:
                plt.ion()
                fig, ax = plt.subplots()
                plot = ax.scatter([], [])
                # set boundaries based on the data
                ax.set_xlim(-2, 2)
                ax.set_ylim(-2, 2)
                ax.set_title(machine)
                for indice in indice_info:
                    if machine in indice_info[indice]:
                        ax.scatter(data_1[indice], data_2[indice], color=machine_colors[machine])

        return ax

    def voronoi(self, X_test=None, y_test=None, line_points=200, epsilon=None, indice_info=None, MSE=None, plot_machines=["ridge", "lasso", "random_forest", "tree"], machine_colors=None, gradient=False, single=False):
        """
        Plot the input space and color query points as a Voronoi Tesselation based on the optimal machine used for that point.

        Parameters
        ----------
        epsilon: float, optional
            Epsilon value to use for diagnostics. Used to find indice_info if it isn't passed.
     
        line_points: int, optional
            if epsilon is not passed, optimal epsilon is found per point. Used to find indice_info if it isn't passed.

        indice_info: dicitonary, optional
            dictionary mapping indice to optimal machines.
      
        MSE: dictionary, optional
            dictionary mapping indice to mean-squared error for optimal machines

        plot_machines: list, optional
            list of machines to be plotted.

        single: bool, optional
            plots a single plot with each machine combination.

        gradient: bool, optional
            instead of aggregating optimal machines, plots a colored plot for each machine,
            shaded according to the mean-squared error of that "region"

        machine_colors: dictionary, optional
            Depending on the kind of coloring, a dictionary mapping machines to colors.
        """

        if X_test is None:
            X_test = self.X_test
        if y_test is None:
            y_test = self.y_test

        if indice_info is None:
            indice_info, MSE = self.indice_info(line_points, epsilon)

        # passing input space to set up voronoi regions.
        points = np.hstack((np.reshape(X_test[:, 0], (len(X_test[:, 0]), 1)), np.reshape(X_test[:, 1], (len(X_test[:, 1]), 1))))
        vor = Voronoi(points)
        # use helper Voronoi
        regions, vertices = voronoi_finite_polygons_2d(vor)

        # # colorize
        if not single:
            for machine in plot_machines:
                fig, ax = plt.subplots()
                plot = ax.scatter([], [])
                ax.set_title(machine)
                indice = 0
                for region in regions:
                    ax.plot(X_test[:, 0][indice], X_test[:, 1][indice], 'ko')
                    polygon = vertices[region]
                    if gradient is True and MSE is not None:
                        # we find closest index from range to give gradient value
                        mse_range = np.linspace(min(MSE.values()), max(MSE.values()), 10)
                        num = min(mse_range, key=lambda x: abs(x - MSE[indice]))
                        index = np.where(mse_range == num)
                        alpha = index[0][0] / 10.0 + 0.2
                        if alpha > 1.0:
                            alpha = 1.0
                        # we fill the polygon with the appropriate gradient
                        if machine in indice_info[indice]:
                            ax.fill(*zip(*polygon), alpha=alpha, color='r')
                    else:
                        # if it isn't gradient based we just color red or blue depending on whether that point uses the machine in question
                        if machine in indice_info[indice]:
                            ax.fill(*zip(*polygon), alpha=0.4, color='r', label="")
                        else:
                            ax.fill(*zip(*polygon), alpha=0.4, color='b', label="")
                    indice += 1

                ax.axis('equal')
                plt.xlim(vor.min_bound[0] - 0.1, vor.max_bound[0] + 0.1)
                plt.ylim(vor.min_bound[1] - 0.1, vor.max_bound[1] + 0.1)

            return vor

        if single:

            if machine_colors is None:
                machine_colors = gen_machine_colors(indice_info=indice_info)

            fig, ax = plt.subplots()
            plot = ax.scatter([], [])
            ax.set_title("All Machines")
            indice = 0
            for region in regions:

                ax.plot(X_test[:, 0][indice], X_test[:, 1][indice], 'ko')
                polygon = vertices[region]
                ax.fill(*zip(*polygon), alpha=0.2, color=machine_colors[indice_info[indice]], label=create_labels(indice_info[indice]))
                indice += 1

            ax.axis('equal')
            plt.xlim(vor.min_bound[0] - 0.1, vor.max_bound[0] + 0.1)
            plt.ylim(vor.min_bound[1] - 0.1, vor.max_bound[1] + 0.1)
            try:
                plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
            except ValueError:
                return vor
            return vor



# Licensed under the MIT License - https://opensource.org/licenses/MIT

from sklearn import linear_model
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.svm import SVR

from sklearn.utils import shuffle
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import GridSearchCV

import matplotlib.pyplot as plt
import math
import numpy as np
import random
import logging

logger = logging.getLogger('pycobra.ewa')


class Ewa(BaseEstimator):
    """
    Exponential Weighted Average aggregation method.
    Implementation based on work by:
    M. Mojirsheibani (1999), Combining Classifiers via Discretization,
    Journal of the American Statistical Association.

    Parameters
    ----------
    random_state: integer or a numpy.random.RandomState object.
        Set the state of the random number generator to pass on to shuffle and loading machines, to ensure
        reproducibility of your experiments, for example.

    beta: float, optional
        Parameter to be passed when creating machine weights for EWA.

    Attributes
    ----------
    machines_ : dictionary
        dictionary mapping name of machine to the object.

    machine_predictions_: A dictionary which maps machine name to it's predictions over X_l
            This value is used to determine which points from y_l are used to aggregate.
    """
    def __init__(self, random_state=None, beta=None):
        self.random_state = random_state
        self.beta = beta

    def fit(self, X, y, default=True, X_k=None, X_l=None, y_k=None, y_l=None):
        """
        Parameters
        ----------
        X: array-like, [n_features]
            Training data which will be used to create the EWA aggregate.

        y: array-like, shape = [n_samples]
            Target values used to train the machines used in the EWA aggregate.

        default: bool, optional
            If set as true then sets up EWA with default machines and splitting.

        X_k : shape = [n_samples, n_features], optional
            Training data which is used to train the machines loaded into Ewa.
            Can be loaded directly into EWA; if not, the split_data method is used as default.

        y_k : array-like, shape = [n_samples], optional
            Target values used to train the machines loaded into EWA.

        X_l : shape = [n_samples, n_features], optional
            Training data which is used during the aggregation of EWA.
            Can be loaded directly into EWA; if not, the split_data method is used as default.

        y_l : array-like, shape = [n_samples], optional
            Target values which are actually used in the aggregation of EWA.

        Returns
        -------
        self : returns an instance of self.

        Notes
        -----

        We store the data used to train the machines because this information is used to make the prediction.

        """

        X, y = check_X_y(X, y)
        self.X_ = X
        self.y_ = y
        self.X_k_ = X_k
        self.X_l_ = X_l
        self.y_k_ = y_k
        self.y_l_ = y_l
        self.estimators_ = {}

        # set-up Ewa with default machines
        if default:
            self.split_data()
            self.load_default()
            self.load_machine_weights(self.beta)

        return self


    def set_beta(self, X_beta=None, y_beta=None, betas=None):
        """
        Parameters
        ----------

        betas: list, optional
            List of betas to find optimal beta for weights

        X_beta : shape = [n_samples, n_features]
            Used if no beta is passed to find the optimal beta for data passed.

        y_beta : array-like, shape = [n_samples]
            Used if no beta is passed to find the optimal beta for data passed.
        """

        if self.beta is None and X_beta is not None:
            if betas is None:
                betas = [0.001, 0.01, 0.1,  1.0, 10.0, 100.0]
            tuned_parameters = [{'beta': betas}]
            clf = GridSearchCV(self, tuned_parameters, cv=5, scoring="neg_mean_squared_error")
            clf.fit(X_beta, y_beta)
            self.beta = clf.best_params_["beta"]


    def split_data(self, k=None, l=None, shuffle_data=False):
        """
        Split the data into different parts for training machines and for aggregation.

        Parameters
        ----------
        k : int, optional
            k is the number of points used to train the machines.
            Those are the first k points of the data provided.

        l: int, optional
            l is the number of points used to form the EWA aggregate.

        shuffle: bool, optional
            Boolean value to decide to shuffle the data before splitting.

        Returns
        -------
        self : returns an instance of self.
        """

        if shuffle_data:
            self.X_, self.y_ = shuffle(self.X_, self.y_, random_state=self.random_state)

        if k is None and l is None:
            k = int(len(self.X_) / 2)
            l = int(len(self.X_))

        if k is not None and l is None:
            l = len(self.X_) - k

        if l is not None and k is None:
            k = len(self.X_) - l

        self.X_k_ = self.X_[:k]
        self.X_l_ = self.X_[k:l]
        self.y_k_ = self.y_[:k]
        self.y_l_ = self.y_[k:l]

        return self


    def load_default(self, machine_list=['lasso', 'tree', 'ridge', 'random_forest', 'svm']):
        """
        Loads 4 different scikit-learn regressors by default.

        Parameters
        ----------
        machine_list: optional, list of strings
            List of default machine names to be loaded.

        """
        for machine in machine_list:
            try:
                if machine == 'lasso':
                    self.estimators_['lasso'] = linear_model.LassoCV(random_state=self.random_state).fit(self.X_k_, self.y_k_)
                if machine == 'tree':
                    self.estimators_['tree'] = DecisionTreeRegressor(random_state=self.random_state).fit(self.X_k_, self.y_k_)
                if machine == 'ridge':
                    self.estimators_['ridge'] = linear_model.RidgeCV().fit(self.X_k_, self.y_k_)
                if machine == 'random_forest':
                    self.estimators_['random_forest'] = RandomForestRegressor(random_state=self.random_state).fit(self.X_k_, self.y_k_)
                if machine == 'svm':
                    self.estimators_['svm'] = SVR().fit(self.X_k_, self.y_k_)
            except ValueError:
                continue

    def load_machine(self, machine_name, machine):
        """
        Adds a machine to be used during the aggregation strategy.
        The machine object must have been trained using X_k and y_k, and must have a 'predict()' method.
        After the machine is loaded, for it to be used during aggregation, load_machine_predictions must be run.

        Parameters
        ----------
        machine_name : string
            Name of the machine you are loading

        machine: machine/regressor object
            The regressor machine object which is mapped to the machine_name

        Returns
        -------
        self : returns an instance of self.
        """

        self.estimators_[machine_name] = machine
        return self


    def load_machine_weights(self, beta=None):
        """
        Loads the EWA weights for each machine based on the training data.
        Should be run after all the machines to be used for aggregation is loaded.

        Parameters
        ----------
        beta : float
            Inverse temperature parameter to form the weights.

        Returns
        -------
        self : returns an instance of self.
        """

        if self.beta is None and beta is None:
            beta = 1.0
        if self.beta is not None and beta is None:
            beta = self.beta

        self.machine_MSE_ = {}
        self.machine_weight_ = {}
        for machine in self.estimators_:
            self.machine_MSE_[machine] = mean_squared_error(self.y_l_, self.estimators_[machine].predict(self.X_l_))
            self.machine_weight_[machine] = np.exp(beta * self.machine_MSE_[machine])
            if self.machine_weight_[machine] == np.inf:
                logger.info("MSE too high, setting equal weights to all machines")
                for machine in self.estimators_:
                    self.machine_weight_[machine] = 1 / len(self.estimators_)
                return self

        normalise = sum(self.machine_weight_.values(), 0.0)
        self.machine_weight_ = {k: v / normalise for k, v in self.machine_weight_.items()}

        return self


    def predict(self, X):
        """
        Parameters
        ----------
        X: array-like, [n_features]

        Returns
        -------
        result: returns prediction
        """
        X = check_array(X)

        result = 0.0
        for machine in self.estimators_:
            result += self.machine_weight_[machine] * self.estimators_[machine].predict(X)

        return result


    def plot_machine_weights(self, figsize=8):
        """
        Plot each machine weights

        Parameteres
        -----------
        figsize: float, optional
            Size of plot.
        """

        plt.bar(range(len(self.machine_weight_)), self.machine_weight_.values(), align='center')
        plt.xticks(range(len(self.machine_weight_)), self.machine_weight_.keys())

        plt.show()



# Licensed under the MIT License - https://opensource.org/licenses/MIT
from sklearn.metrics import mean_squared_error, accuracy_score

import itertools
import numpy as np
from pycobra.cobra import Cobra
from pycobra.ewa import Ewa
from pycobra.classifiercobra import ClassifierCobra
from pycobra.kernelcobra import KernelCobra

import math
import logging

logger = logging.getLogger('pycobra.diagnostics')


class Diagnostics():
    """
    Optimization of hyperparameters, and error details.
    """
    def __init__(self, aggregate, X_test=None, y_test=None, load_MSE=False, random_state=None):
        """
        Parameters
        ----------
        aggregate: pycobra.cobra.Cobra or pycobra.cobra.Ewa object
            aggregate on which we want to run our analysis on.

        X_test : array-like, shape = [n_samples, n_features], optional.
            Testing data.

        y_test : array-like, shape = [n_samples], optional.
            Test data target values.

        load_MSE: bool, optional
            loads MSE and error bound values into diagnostics object.

        random_state: integer or a numpy.random.RandomState object.
            Set the state of the random number generator to pass on to shuffle and loading machines, to ensure
            reproducibility of your experiments, for example.
        """
        self.aggregate = aggregate
        # X_test and y_test is only to compare MSEs of machines.
        self.X_test = X_test
        self.y_test = y_test

        if load_MSE and X_test is not None and type(self.aggregate) is not ClassifierCobra:
            self.load_MSE()

        if type(self.aggregate) is ClassifierCobra and load_MSE:
            self.load_errors()

        self.random_state = random_state
        if self.random_state is None:
            self.random_state = self.aggregate.random_state


    def load_MSE(self):
        """
        Computes MSE and error bound for each Machine based on test data.

        Returns
        -------
        self : returns an instance of self.

        """
        self.machine_test_results = {}
        self.machine_MSE = {}

       
        names_dict = {Cobra: "Cobra", Ewa: "EWA"}

        for name in names_dict:
            if type(self.aggregate) is name:
                self.machine_test_results[names_dict[name]] = self.aggregate.predict(self.X_test)
                self.machine_MSE[names_dict[name]] = mean_squared_error(self.y_test, self.machine_test_results[names_dict[name]])            


        for machine in self.aggregate.estimators_:
            self.machine_test_results[machine] = self.aggregate.estimators_[machine].predict(self.X_test)
            # add MSE
            self.machine_MSE[machine] = mean_squared_error(self.y_test, self.machine_test_results[machine])

        # COBRA bound error
        power = - 2 / (len(self.aggregate.estimators_) + 2)
        self.error_bound = math.pow(len(self.aggregate.X_l_), power)

        return self

    def load_errors(self):
        """
        Computes accuracy score for each Machine based on test data.

        Returns
        -------
        self : returns an instance of self.

        """
        self.machine_test_results = {}
        self.machine_error = {}

        self.machine_test_results["ClassifierCobra"] = self.aggregate.predict(self.X_test)
        self.machine_error["ClassifierCobra"] = 1 - accuracy_score(self.y_test, self.machine_test_results["ClassifierCobra"])

        for machine in self.aggregate.estimators_:
            self.machine_test_results[machine] = self.aggregate.estimators_[machine].predict(self.X_test)
            # add MSE
            self.machine_error[machine] = 1 - accuracy_score(self.y_test, self.machine_test_results[machine])

    def optimal_alpha(self, X, y, single=False, epsilon=None, info=False):
        """
        Find the optimal alpha for testing data for the COBRA predictor.

        Parameteres
        -----------

        X: array-like, [n_features]
            Vector for which we want optimal alpha values

        y: float
            Target value for query to compare.

        single: boolean, optional
            Option to calculate optimal alpha for a single query point instead.

        info: bool, optional
            Returns MSE dictionary for each alpha value

        epsilon: float, optional
            fixed epsilon value to help determine optimal alpha.

        Returns
        -------

        MSE: dictionary mapping alpha with mean squared errors
        opt: optimal alpha combination

        """
        if epsilon is None:
            epsilon = self.aggregate.epsilon

        MSE = {}
        for alpha in range(1, len(self.aggregate.estimators_) + 1):
            machine = Cobra(random_state=self.random_state, epsilon=epsilon)
            machine.fit(self.aggregate.X_, self.aggregate.y_)
            # for a single data point
            if single:
                result = machine.predict(X, alpha=alpha)
                MSE[alpha] = np.square(y - result)
            else:
                results = machine.predict(X, alpha=alpha)
                MSE[alpha] = (mean_squared_error(y, results))

        if info:
            return MSE
        opt = min(MSE, key=MSE.get)
        return opt, MSE[opt]


    def optimal_machines(self, X, y, single=False, epsilon=None, info=False):
        """
        Find the optimal combination of machines for testing data for the COBRA predictor.

        Parameteres
        -----------

        X: array-like, [n_features]
            Vector for which we want optimal machine combinations.

        y: float
            Target value for query to compare.

        single: boolean, optional
            Option to calculate optimal machine combinations for a single query point instead.

        info: bool, optional
            Returns MSE dictionary for each machine combination value

        epsilon: float, optional
            fixed epsilon value to help determine optimal machines.

        Returns
        -------

        MSE: dictionary mapping machines with mean squared errors
        opt: optimal machines combination

        """
        if epsilon is None:
            epsilon = self.aggregate.epsilon

        n_machines = np.arange(1, len(self.aggregate.estimators_) + 1)
        MSE = {}
        for num in n_machines:
            machine_names = self.aggregate.estimators_.keys()
            use = list(itertools.combinations(machine_names, num))
            for combination in use:
                machine = Cobra(random_state=self.random_state, epsilon=epsilon)
                machine.fit(self.aggregate.X_, self.aggregate.y_, default=False)
                machine.split_data()
                machine.load_default(machine_list=combination)
                machine.load_machine_predictions()
                if single:
                    result = machine.predict(X.reshape(1, -1))
                    MSE[combination] = np.square(y - result)
                else:
                    results = machine.predict(X)
                    MSE[combination] = (mean_squared_error(y, results))

        if info:
            return MSE
        opt = min(MSE, key=MSE.get)
        return opt, MSE[opt]


    def optimal_epsilon(self, X, y, line_points=200, info=False):
        """
        Find the optimal epsilon value for the COBRA predictor.

        Parameteres
        -----------

        X: array-like, [n_features]
            Vector for which we want for optimal epsilon.

        y: float
            Target value for query to compare.

        line_points: integer, optional
            Number of epsilon values to traverse the grid.

        info: bool, optional
            Returns MSE dictionary for each epsilon value.

        Returns
        -------

        MSE: dictionary mapping epsilon with mean squared errors
        opt: optimal epsilon value

        """

        a, size = sorted(self.aggregate.all_predictions_), len(self.aggregate.all_predictions_)
        res = [a[i + 1] - a[i] for i in range(size) if i+1 < size]
        emin = min(res)
        emax = max(a) - min(a)
        erange = np.linspace(emin, emax, line_points)

        MSE = {}
        for epsilon in erange:
            machine = Cobra(random_state=self.random_state, epsilon=epsilon)
            machine.fit(self.aggregate.X_, self.aggregate.y_)
            results = machine.predict(X)
            MSE[epsilon] = (mean_squared_error(y, results))

        if info:
            return MSE
        opt = min(MSE, key=MSE.get)
        return opt, MSE[opt]


    def optimal_split(self, X, y, split=None, epsilon=None, info=False, graph=False):

        """
        Find the optimal combination split (D_k, D_l) for fixed epsilon value for the COBRA predictor.

        Parameteres
        -----------

        X: array-like, [n_features]
            Vector for which we want for optimal split.

        y: float
            Target value for query to compare.

        epsilon: float, optional.
            fixed epsilon value to help determine optimal machines.

        split: list, optional.
            D_k, D_l break-up to calculate MSE

        info: bool, optional.
            Returns MSE dictionary for each split.

        graph: bool, optional.
            Plots graph of MSE vs split

        Returns
        -------

        MSE: dictionary mapping split with mean squared errors
        opt: optimal epsilon value

        """
        if epsilon is None:
            epsilon = self.aggregate.epsilon

        if split is None:
            split = [(0.20, 0.80), (0.40, 0.60), (0.50, 0.50), (0.60, 0.40), (0.80, 0.20)]

        MSE = {}
        for k, l in split:
            machine = Cobra(random_state=self.random_state, epsilon=epsilon)
            machine.fit(self.aggregate.X_, self.aggregate.y_, default=False)
            machine.split_data(int(k * len(self.aggregate.X_)), int((k + l) * len(self.aggregate.X_)))
            machine.load_default()
            machine.load_machine_predictions()
            results = machine.predict(X)
            MSE[(k, l)] = (mean_squared_error(y, results))

        if graph:
            import matplotlib.pyplot as plt
            ratio, mse = [], []
            for value in split:
                ratio.append(value[0])
                mse.append(MSE[value])
            plt.plot(ratio, mse)

        if info:
            return MSE
        opt = min(MSE, key=MSE.get)
        return opt, MSE[opt]


    def optimal_alpha_grid(self, X, y, line_points=200, info=False):
        """
        Find the optimal epsilon and alpha for a single query point for the COBRA predictor.

        Parameteres
        -----------

        X: array-like, [n_features]
            Vector for which we want optimal alpha and epsilon values

        y: float
            Target value for query to compare.

        line_points: integer, optional
            Number of epsilon values to traverse the grid.

        info: bool, optional
            Returns MSE dictionary for each epsilon/alpha value

        Returns
        -------

        MSE: dictionary mapping (alpha, epsilon) with mean squared errors
        opt: optimal epislon/alpha combination

        """

        # code to find maximum and minimum distance between predictions to create grid
        a, size = sorted(self.aggregate.all_predictions_), len(self.aggregate.all_predictions_)
        res = [a[i + 1] - a[i] for i in range(size) if i+1 < size]
        emin = min(res)
        emax = max(a) - min(a)
        erange = np.linspace(emin, emax, line_points)
        n_machines = np.arange(1, len(self.aggregate.estimators_) + 1)
        MSE = {}

        # looping over epsilon and alpha values
        for epsilon in erange:
            for num in n_machines:
                machine = Cobra(random_state=self.random_state, epsilon=epsilon)
                machine.fit(self.aggregate.X_, self.aggregate.y_)
                result = machine.predict(X.reshape(1, -1), alpha=num)
                MSE[(num, epsilon)] = np.square(y - result)

        if info:
            return MSE

        opt = min(MSE, key=MSE.get)
        return opt, MSE[opt]


    def optimal_machines_grid(self, X, y, line_points=200, info=False):
        """
        Find the optimal epsilon and machine-combination for a single query point for the COBRA predictor.

        Parameteres
        -----------

        X: array-like, [n_features]
            Vector for which we want optimal machines and epsilon values

        y: float
            Target value for query to compare.

        line_points: integer, optional
            Number of epsilon values to traverse the grid.

        info: bool, optional
            Returns MSE dictionary for each epsilon/machine value.

        Returns
        -------

        MSE: dictionary mapping (machine combination, epsilon) with mean squared errors
        opt: optimal epislon/machine combination

        """

        # code to find maximum and minimum distance between predictions to create grid
        a, size = sorted(self.aggregate.all_predictions_), len(self.aggregate.all_predictions_)
        res = [a[i + 1] - a[i] for i in range(size) if i+1 < size]
        emin = min(res)
        emax = max(a) - min(a)
        erange = np.linspace(emin, emax, line_points)
        n_machines = np.arange(1, len(self.aggregate.estimators_) + 1)
        MSE = {}

        for epsilon in erange:
            for num in n_machines:
                machine_names = self.aggregate.estimators_.keys()
                use = list(itertools.combinations(machine_names, num))
                for combination in use:
                    machine = Cobra(random_state=self.random_state, epsilon=epsilon)
                    machine.fit(self.aggregate.X_, self.aggregate.y_, default=False)
                    machine.split_data()
                    machine.load_default(machine_list=combination)
                    machine.load_machine_predictions()
                    result = machine.predict(X.reshape(1, -1))
                    MSE[(combination, epsilon)] = np.square(y - result)

        if info:
            return MSE
        opt = min(MSE, key=MSE.get)
        return opt, MSE[opt]


    def optimal_beta(self, X, y, betas=None, info=False):
        """
        Find the optimal beta value for the Ewa predictor.

        Parameteres
        -----------

        X: array-like, [n_features]
            Vector for which we want for optimal beta.

        y: float
            Target value for query to compare.

        betas: list, optional
            List of beta values to iterate over for optimal beta.

        info: bool, optional
            Returns MSE dictionary for each beta value.

        Returns
        -------

        MSE: dictionary mapping epsilon with mean squared errors
        opt: optimal beta value

        """

        if betas is None:
            betas = np.arange(0.1, 10)

        MSE = {}
        for beta in betas:
            machine = Ewa(random_state=self.random_state, beta=beta)
            machine.fit(self.aggregate.X_, self.aggregate.y_)
            results = machine.predict(X)
            MSE[beta] = (mean_squared_error(y, results))

        if info:
            return MSE
        opt = min(MSE, key=MSE.get)
        return opt, MSE[opt]


    def optimal_kernelbandwidth(self, X, y, bandwidths=None, info=False):
        """
        Find the optimal bandwidth value for the KernelCobra predictor.

        Parameteres
        -----------

        X: array-like, [n_features]
            Vector for which we want for optimal bandwidths.

        y: float
            Target value for query to compare.

        bandwidths: list, optional
            List of bandwidth values to iterate over for optimal bandwidth.

        info: bool, optional
            Returns MSE dictionary for each bandwidth value.

        Returns
        -------

        MSE: dictionary mapping epsilon with mean squared errors
        opt: optimal bandwidth value

        """

        if bandwidths is None:
            bandwidths = np.arange(0.05, 2.55, step=0.10)

        MSE = {}
        for bandwidth in bandwidths:
            machine = KernelCobra(random_state=self.random_state)
            machine.fit(self.aggregate.X_, self.aggregate.y_)
            results = machine.predict(X, bandwidth=bandwidth)
            MSE[bandwidth] = (mean_squared_error(y, results))

        if info:
            return MSE
        opt = min(MSE, key=MSE.get)
        return opt, MSE[opt]



# Licensed under the MIT License - https://opensource.org/licenses/MIT

from sklearn import linear_model
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.svm import SVR
from sklearn.neural_network import MLPRegressor

from sklearn.utils import shuffle
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.model_selection import GridSearchCV

import math
import numpy as np
import random
import logging
import numbers


logger = logging.getLogger('pycobra.kernelcobra')


class KernelCobra(BaseEstimator):
    """
    Regression algorithm as introduced by
    Kernel-COBRA: A combined regression-classification strategy using Kernels.
    Based on the paper by Guedj, Srinivasa Desikan [2018].

    Parameters
    ----------
    random_state: integer or a numpy.random.RandomState object.
        Set the state of the random number generator to pass on to shuffle and loading machines, to ensure
        reproducibility of your experiments, for example.


    Attributes
    ----------
    estimators_: A dictionary which maps machine names to the machine objects.
            The machine object must have a predict method for it to be used during aggregation.

    machine_predictions_: A dictionary which maps machine name to it's predictions over X_l
            This value is used to determine which points from y_l are used to aggregate.

    all_predictions_: numpy array with all the predictions, to be used for bandwidth manipulation.

    """

    def __init__(self, random_state=None, machine_list='basic'):
        self.random_state = random_state
        self.machine_list = machine_list

    def fit(self, X, y, default=True, X_k=None, X_l=None, y_k=None, y_l=None):
        """
        Parameters
        ----------
        X: array-like, [n_samples, n_features]
            Training data which will be used to create the COBRA aggregate.

        y: array-like, shape = [n_samples]
            Target values used to train the machines used in the aggregation.

        default: bool, optional
            If set as true then sets up COBRA with default machines and splitting.

        X_k : shape = [n_samples, n_features]
            Training data which is used to train the machines used in the aggregation.
            Can be loaded directly into COBRA; if not, the split_data method is used as default.

        y_k : array-like, shape = [n_samples]
            Target values used to train the machines used in the aggregation.

        X_l : shape = [n_samples, n_features]
            Training data which is used to form the aggregate.
            Can be loaded directly into COBRA; if not, the split_data method is used as default.

        y_l : array-like, shape = [n_samples]
            Target values which are actually used to form the aggregate.
        """

        X, y = check_X_y(X, y)
        self.X_ = X
        self.y_ = y
        self.X_k_ = X_k
        self.X_l_ = X_l
        self.y_k_ = y_k
        self.y_l_ = y_l
        self.estimators_ = {}
        # set-up COBRA with default machines
        if default:
            self.split_data()
            self.load_default(machine_list=self.machine_list)
            self.load_machine_predictions()

        return self


    def pred(self, X, kernel=None, metric=None, bandwidth=1, **kwargs):
        """
        Performs the Kernel-COBRA aggregation scheme, used in predict method.

        Parameters
        ----------
        X: array-like, [n_features]

        kernel: function, optional
            kernel refers to the kernel method which we wish to use to perform the aggregation.

        metric: function, optional
            metric refers to the metric method which we wish to use to perform the aggregation.

        bandwidth: float, optional
            Bandwidth for the deafult kernel value (gaussian), and is set to 1.

        kwargs requires you to pass arguments with "kernel_params" and "metric_params", if the custom kernel or metric
        has more paramteres.        

        Returns
        -------
        avg: prediction

        """

        a = np.zeros(len(self.X_l_))
        for machine in self.estimators_:
            val = self.estimators_[machine].predict(X)
            for index, value in np.ndenumerate(self.machine_predictions_[machine]):
                if metric is not None:
                    try:
                        a[index] += metric(value, val, kwargs["metric_params"])
                    except KeyError:
                        a[index] += metric(value, val)
                else:
                    a[index] += math.fabs(value - val)

        # normalise the array
        if kernel is not None:
            try:
                a = np.divide(kernel(a, kwargs["kernel_params"]), np.sum(kernel(a, kwargs["kernel_params"])))
            except KeyError:
                a = np.divide(kernel(a), np.sum(kernel(a)))
        else:
            exp = np.nan_to_num(np.exp(- bandwidth * a))
            a = np.nan_to_num(np.divide(exp, np.sum(exp)))

        return np.sum(np.multiply(self.y_l_, a))


    def predict(self, X, kernel=None, metric=None, bandwidth=1, **kwargs):
        """
        Performs the Kernel-COBRA aggregation scheme, calls pred.

        Parameters
        ----------
        X: array-like, [n_features]

        kernel: function, optional
            kernel refers to the kernel method which we wish to use to perform the aggregation.

        metric: function, optional
            metric refers to the metric method which we wish to use to perform the aggregation.

        bandwidth: float, optional
            Bandwidth for the deafult kernel value (gaussian), and is set to 1.

        kwargs requires you to pass arguments with "kernel_params" and "metric_params", if the custom kernel or metric
        has more paramteres.        

        Returns
        -------
        avg: prediction

        """

        X = check_array(X)

        if X.ndim == 1:
            return self.pred(X.reshape(1, -1))

        result = np.zeros(len(X))
        avg_points = 0
        index = 0
        for vector in X:
            result[index] = self.pred(vector.reshape(1, -1), kernel=kernel, metric=metric, bandwidth=bandwidth, **kwargs)
            index += 1

        return result


    def split_data(self, k=None, l=None, shuffle_data=False):
        """
        Split the data into different parts for training machines and for aggregation.

        Parameters
        ----------
        k : int, optional
            k is the number of points used to train the machines.
            Those are the first k points of the data provided.

        l: int, optional
            l is the number of points used to form the COBRA aggregate.

        shuffle: bool, optional
            Boolean value to decide to shuffle the data before splitting.

        Returns
        -------
        self : returns an instance of self.
        """

        if shuffle_data:
            self.X_, self.y_ = shuffle(self.X_, self.y_, random_state=self.random_state)

        if k is None and l is None:
            k = int(len(self.X_) / 2)
            l = int(len(self.X_))

        if k is not None and l is None:
            l = len(self.X_) - k

        if l is not None and k is None:
            k = len(self.X_) - l

        self.X_k_ = self.X_[:k]
        self.X_l_ = self.X_[k:l]
        self.y_k_ = self.y_[:k]
        self.y_l_ = self.y_[k:l]

        return self


    def load_default(self, machine_list='basic'):
        """
        Loads 4 different scikit-learn regressors by default. The advanced list adds more machines. 
        Parameters
        ----------
        machine_list: optional, list of strings
            List of default machine names to be loaded. 
            Default is basic,
        Returns
        -------
        self : returns an instance of self.
        """
        if machine_list == 'basic':
            machine_list = ['tree', 'ridge', 'random_forest', 'svm']
        if machine_list == 'advanced':
            machine_list=['lasso', 'tree', 'ridge', 'random_forest', 'svm', 'bayesian_ridge', 'sgd']

        self.estimators_ = {}
        for machine in machine_list:
            try:
                if machine == 'lasso':
                    self.estimators_['lasso'] = linear_model.LassoCV(random_state=self.random_state).fit(self.X_k_, self.y_k_)
                if machine == 'tree':
                    self.estimators_['tree'] = DecisionTreeRegressor(random_state=self.random_state).fit(self.X_k_, self.y_k_)
                if machine == 'ridge':
                    self.estimators_['ridge'] = linear_model.RidgeCV().fit(self.X_k_, self.y_k_)
                if machine == 'random_forest':
                    self.estimators_['random_forest'] = RandomForestRegressor(random_state=self.random_state).fit(self.X_k_, self.y_k_)
                if machine == 'svm':
                    self.estimators_['svm'] = SVR().fit(self.X_k_, self.y_k_)
                if machine == 'sgd':
                    self.estimators_['sgd'] = linear_model.SGDRegressor(random_state=self.random_state).fit(self.X_k_, self.y_k_)
                if machine == 'bayesian_ridge':
                    self.estimators_['bayesian_ridge'] = linear_model.BayesianRidge().fit(self.X_k_, self.y_k_)
            except ValueError:
                continue
        return self


    def load_machine(self, machine_name, machine):
        """
        Adds a machine to be used during the aggregation strategy.
        The machine object must have been trained using X_k and y_k, and must have a 'predict()' method.
        After the machine is loaded, for it to be used during aggregation, load_machine_predictions must be run.

        Parameters
        ----------
        machine_name : string
            Name of the machine you are loading

        machine: machine/regressor object
            The regressor machine object which is mapped to the machine_name

        Returns
        -------
        self : returns an instance of self.
        """

        self.estimators_[machine_name] = machine

        return self


    def load_machine_predictions(self, predictions=None):
        """
        Stores the trained machines' predicitons on training data in a dictionary, to be used for predictions.
        Should be run after all the machines to be used for aggregation is loaded.

        Parameters
        ----------
        predictions: dictionary, optional
            A pre-existing machine:predictions dictionary can also be loaded.

        Returns
        -------
        self : returns an instance of self.
        """
        self.machine_predictions_ = {}
        self.all_predictions_ = np.array([])
        if predictions is None:
            for machine in self.estimators_:
                self.machine_predictions_[machine] = self.estimators_[machine].predict(self.X_l_)
                # all_predictions_ is used in the diagnostics class, and for initialising epsilon
                self.all_predictions_ = np.append(self.all_predictions_, self.machine_predictions_[machine])

        if predictions is not None:
            self.machine_predictions_ = predictions

        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under the MIT License - https://opensource.org/licenses/MIT

from setuptools import setup

DESCRIPTION = 'A set of python modules for ensemble learning and visualisation'
with open('README.rst') as f:
    LONG_DESCRIPTION = f.read()

version = "0.2.4"
setup(name='pycobra',
      version=version,
      description='Python implementation of COBRA algorithm with regression analysis',
      author=['Bhargav Srinivasa Desikan', 'Benjamin Guedj'],
      author_email=['bhargavvader@gmail.com', 'benjamin.guedj@inria.fr'],
      url='https://github.com/bhargavvader/pycobra',
      license='MIT',
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Operating System :: OS Independent',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering'
      ],
      packages=['pycobra'],
      install_requires=[
          'numpy',
          'scipy',
          'scikit-learn',
          'matplotlib',
          'seaborn',
          'pandas'

      ],
      test_suite='tests',
      keywords=[
          'Aggregation of Predictors',
          'Regression Analysis',
          'Voronoi Tesselation',
          'Statistical Aggregation'
      ],
      long_description=LONG_DESCRIPTION)
